/*\
||| Earendil's GD module for uLPC
||| Began Thu Jul 17 1996
|||
\*/
#include "gd1.2/gd.h"
#include "gd1.2/gdfontg.h"
#include "gd1.2/gdfontl.h"
#include "gd1.2/gdfontmb.h"
#include "gd1.2/gdfonts.h"
#include "gd1.2/gdfontt.h"

#include <string.h>
#include <stdio.h>

#include "array.h"
#include "global.h"
#include "types.h"
#include "interpret.h"
#include "add_efun.h"
#include "svalue.h"
#include "stralloc.h"
#include "object.h"

struct gd_image_object {
  gdImagePtr im;
  int color;
};

#define THIS ((struct gd_image_object *)(fp->current_storage))

static struct program *gd_program;


void init_gd_callback(char *foo, struct object *o) {
}

void exit_gd_callback(char *foo, struct object *o) {
}

void check_valid_color(int c, char *func) {
  char *s;

  if( c < 0 || c + 1 > THIS->im->colorsTotal ) {
    s = (char *) malloc(80);
    sprintf(s, "Unallocated color (%d) passed to gd->%s()\n",
	c, func);
    error(s);
  }

  if( THIS->color == -1 )
    error("No color(s) allocated yet in gd image\n");
}
   
/*
 ** Local functions
 */
static void gd_create(INT32 args) {
  int x, y;

  if( args ) {
    if( args != 2 )
      error("Incorrect number of arguments to gd->create()\n");
    if( sp[1-args].type != T_INT )
      error("Bad argument 2 to gd->create()\n");
    if( sp[-args].type != T_INT )
      error("Bad argument 1 to gd->create()\n");
    x = sp[-args].u.integer;
    y = sp[1-args].u.integer;

    THIS->im = gdImageCreate(x,y);
    pop_n_elems(args);
  } else
    THIS->im = NULL;

  THIS->color = -1;

}

static void gd_alloc_color(INT32 args) {
  int value, red, green, blue;

  if( args != 3 )
    error("Incorrect number of arguments to gd->alloc_color()\n");
  if( sp[2-args].type != T_INT )
    error("Bad argument 3 to gd->alloc_color()\n");
  if( sp[1-args].type != T_INT )
    error("Bad argument 2 to gd->alloc_color()\n");
  if( sp[-args].type != T_INT )
    error("Bad argument 1 to gd->alloc_color()\n");

  if( THIS->im == NULL )
    error("No Image Open or Created");

  red = sp[-args].u.integer;
  blue = sp[1-args].u.integer;
  green = sp[2-args].u.integer;
  
  value = gdImageColorAllocate(THIS->im, red, blue, green);
  THIS->color = value;

  pop_n_elems(args);
  push_int(value);
}

static void gd_line(INT32 args) {
  int color, x1, y1, x2, y2;

  if( args < 4 || args > 5 )
    error("Incorrect number of arguments to gd->line()\n");

  if( sp[3-args].type != T_INT )
    error("Bad argument 4 to gd->line()\n");

  if( sp[2-args].type != T_INT )
    error("Bad argument 3 to gd->line()\n");

  if( sp[1-args].type != T_INT )
    error("Bad argument 2 to gd->line()\n");

  if( sp[-args].type != T_INT )
    error("Bad argument 1 to gd->line()\n");

  if( args == 5 ) {
    if( sp[4-args].type != T_INT )
      error("Bad argument 5 to gd->line()\n");
    color = sp[4-args].u.integer;
  } else
    color = THIS->color;

  check_valid_color(color, "line");
  THIS->color = color;

  if( THIS->im == NULL )
    error("No Image Open or Created");

  x1 = sp[-args].u.integer;
  y1 = sp[1-args].u.integer;
  x2 = sp[2-args].u.integer;
  y2 = sp[3-args].u.integer;

  gdImageLine(THIS->im, x1, y1, x2, y2, color);

  pop_n_elems(args);
}

static void gd_print_image(INT32 args) {
  if( args )
    error("Incorrect number of arguments to gd->print_image()\n");

  if( THIS->im == NULL )
    error("No Image Open or Created");

  gdImageGif(THIS->im, stdout);
}

static void gd_create_image(INT32 args) {
  int x, y;

  if( args != 2 )
    error("Incorrect number of arguments to gd->create_image()\n");
  if( sp[-args].type != T_INT )
    error("Bad argument 1 to gd->create_image()\n");
  if( sp[1-args].type != T_INT )
    error("Bad argument 2 to gd->create_image()\n");

  x = sp[-args].u.integer;
  y = sp[1-args].u.integer;
  pop_n_elems(args);

  if( THIS->im != NULL )
    gdImageDestroy(THIS->im);

  THIS->im = gdImageCreate(x, y);

  if( THIS->im == NULL )
    push_int(0);
  else 
    push_int(1);
}
  
static void gd_query_size(INT32 args) {
  int x, y, colors;

  if( args )
    error("Too many arguments to gd->query_size()\n");

  if( THIS->im == NULL )
    error("No Image Open or Created");

  x = THIS->im->sx;
  y = THIS->im->sy;
  colors = THIS->im->colorsTotal;

  push_int(x);
  push_int(y);
  push_int(colors);
  f_aggregate(3);
}
  
static void gd_load_image(INT32 args) {
  FILE *f;
  char *fn, *ft;
  int i, sz;

  if( args < 1 )
    error("Too few arguments to gd->load_image()\n");

  if( sp[-args].type != T_STRING )
    error("Bad argument 1 to gd->load_image()\n");
  fn = sp[-args].u.string->str;

  if( args >= 2 ) {
    if( sp[1-args].type != T_STRING )
      error("Bad argument 2 to gd->load_image()\n");
    ft = sp[1-args].u.string->str;
  } else {
    sz = strlen(fn);
  
    for( i=sz-1; i >= 0; i-- )
      if( fn[i] == '.' )
        break;
    ft = (char *) malloc(sz-i+1);
    strcpy(ft, fn+i+1);
  }


  pop_n_elems(args);

  if( !strcasecmp(fn, "stdin") )
    f = stdin;
  else 
    f = fopen(fn, "r");

  if( !f )
    error("Unable to open file in gd->load_image()\n");


  if( THIS->im != NULL )
    gdImageDestroy(THIS->im);


  if( !strcasecmp(ft, "gif") )
    THIS->im = gdImageCreateFromGif(f);
  else if( !strcasecmp(ft, "xbm") )
    THIS->im = gdImageCreateFromXbm(f);
  else if( !strcasecmp(ft, "gd") )
    THIS->im = gdImageCreateFromGd(f);
  else
    error("Unsupported image type in gd->load_image() (Types: gif, xbm, gd)\n");

  fclose(f);

  if( THIS->im == NULL )
    push_int(0);
  else {
    push_int(1);
    THIS->color = THIS->im->colorsTotal - 1;
  }
}

static void gd_box(INT32 args) {
  int x1, y1, x2, y2, color, fill;

  if( args < 4 )
    error("Too few arguments to gd->box()\n");
 

  if( sp[3-args].type != T_INT )
    error("Bad argument 4 to gd->box()\n");

  if( sp[2-args].type != T_INT )
    error("Bad argument 3 to gd->box()\n");

  if( sp[1-args].type != T_INT )
    error("Bad argument 2 to gd->box()\n");

  if( sp[-args].type != T_INT )
    error("Bad argument 1 to gd->box()\n");

  if( THIS->im == NULL )
    error("No Image Open or Created");

  x1 = sp[-args].u.integer;
  y1 = sp[1-args].u.integer;
  x2 = sp[2-args].u.integer;
  y2 = sp[3-args].u.integer;

  if( args >= 5 ) {
    if( sp[4-args].type != T_INT )
      error("Bad argument 4 to gd->box()\n");
    fill = sp[4-args].u.integer;
  } else
    fill = 0;
  if( args >= 6 ) {
    if( sp[5-args].type != T_INT )
      error("Bad argument 5 to gd->box()\n");
    color = sp[5-args].u.integer;
  } else
    color = THIS->color;

  check_valid_color(color, "box");
  THIS->color = color;

  if( fill )
    gdImageFilledRectangle(THIS->im, x1, y1, x2, y2, color);
  else
    gdImageRectangle(THIS->im, x1, y1, x2, y2, color);
}

static void gd_set_transparent(INT32 args) {
  int color;

  if( args ) {
    if( sp[-args].type != T_INT )
      error("Bad argument 1 to gd->set_transparent()\n");
    color = sp[-args].u.integer;
  } else
    color = THIS->color;

  check_valid_color(color, "set_transparent");
  THIS->color = color;

  if( THIS->im == NULL )
    error("No Image Open or Created");

  gdImageColorTransparent(THIS->im, color);
}

static void gd_free_color(INT32 args) {
  int color;

  if( args ) {
    if( sp[-args].type != T_INT )
      error("Bad argument 1 to gd->free_color()\n");
    color = sp[-args].u.integer;
  } else
    color = THIS->color;

  check_valid_color(color, "free_color");

  if( THIS->im == NULL )
    error("No Image Open or Created");

  gdImageColorDeallocate(THIS->im, color);
}

static void gd_query_color(INT32 args) {
  int color, r, g, b;

  if( args ) {
    if( sp[-args].type != T_INT )
      error("Bad Argument 1 to gd->query_color()\n");
    color = sp[-args].u.integer;
  } else
    color = THIS->color;
  
  if( THIS->im == NULL )
    error("No Image Open or Created");

  check_valid_color(color, "query_color");
  
  r = gdImageRed(THIS->im, color);
  g = gdImageGreen(THIS->im, color);
  b = gdImageBlue(THIS->im, color);

  pop_n_elems(args);

  push_int(r);
  push_int(g);
  push_int(b);

  f_aggregate(3);
}

static void gd_fill(INT32 args) {
  int x, y, color;

  if( args < 2 )
    error("Too few arguments to gd->fill()\n");

  if( sp[-args].type != T_INT )
    error("Bad argument 1 to gd->fill()\n");
  if( sp[1-args].type != T_INT )
    error("Bad argument 2 to gd->fill()\n");

  if( THIS->im == NULL )
    error("No Image Open or Created");

  x = sp[-args].u.integer;
  y = sp[1-args].u.integer;

  if( args >= 3 ) {
    if( sp[2-args].type != T_INT )
      error("Bad Argument 3 to gd->fill()\n");
    color = sp[2-args].u.integer;
  } else
    color = THIS->color;

  check_valid_color(color, "fill");
  THIS->color = color;

  pop_n_elems(args);
  gdImageFill(THIS->im, x, y, color);
}

static void gd_query_pixel(INT32 args) {
  int x, y;

  if( args < 2 )
    error("Too few arguments to gd->query_pixel()\n");

  if( sp[-args].type != T_INT )
    error("Bad argument 1 to gd->query_pixel()\n");
  if( sp[1-args].type != T_INT )
    error("Bad argument 2 to gd->query_pixel()\n");

  if( THIS->im == NULL )
    error("No Image Open or Created");

  x = sp[-args].u.integer;
  y = sp[1-args].u.integer;

  pop_n_elems(args);
  push_int(gdImageGetPixel(THIS->im, x, y)); 
}

static void gd_destroy(INT32 args) {
  if( THIS->im == NULL )
    error("No Image Open or Created");

  gdImageDestroy(THIS->im);
  THIS->im = NULL;
  pop_n_elems(args);
}

static void gd_point(INT32 args) {
  int x, y, color;

  if( args < 2 )
    error("Too few arguments to gd->point()\n");
  if( sp[-args].type != T_INT )
    error("Bad argument 1 to gd->point()\n");
  if( sp[1-args].type != T_INT )
    error("Bad argument 2 to gd->point()\n");

  if( THIS->im == NULL )
    error("No Image Open or Created");

  if( args >= 3 ) {
    if( sp[2-args].type != T_INT )
      error("Bad Argument 3 to gd->point()\n");
    color = sp[2-args].u.integer;
  } else
    color = THIS->color;

  check_valid_color(color, "point");
  THIS->color = color;

  x = sp[-args].u.integer;
  y = sp[1-args].u.integer;

  pop_n_elems(args);
  gdImageSetPixel(THIS->im, x, y, color);
}

static void gd_arc(INT32 args) {
  int cx, cy, w, h, s, e, color;

  if( args < 3 )
    error("Too few arguments to gd->arc()\n");

  if( sp[0-args].type != T_INT )
    error("Bad argument 1 to gd->arc()\n");
  cx = sp[0-args].u.integer;

  if( sp[1-args].type != T_INT )
    error("Bad argument 2 to gd->arc()\n");
  cy = sp[1-args].u.integer;
  
  if( sp[2-args].type != T_INT )
    error("Bad argument 3 to gd->arc()\n");
  w = sp[2-args].u.integer;

  if( args >= 4 ) {
    if( sp[3-args].type != T_INT )
      error("Bad argument 4 to gd->arc()\n");
    h = sp[3-args].u.integer;
  } else
    h = w;

  if( args >= 5 ) {
    if( sp[4-args].type != T_INT )
      error("Bad argument 5 to gd->arc()\n");
    s = sp[4-args].u.integer;
  } else
    s = 0;

  if( args >= 6 ) {
    if( sp[5-args].type != T_INT )
      error("Bad argument 6 to gd->arc()\n");
    e = sp[5-args].u.integer;
  } else
    e = 360;

  if( args >= 7 ) {
    if( sp[6-args].type != T_INT )
      error("Bad argument 7 to gd->arc()\n");
    color = sp[6-args].u.integer;
  } else
    color = THIS->color;

  check_valid_color(color, "arc");
  THIS->color = color;

  if( THIS->im == NULL )
    error("No Image Open or Created");

  pop_n_elems(args);

  gdImageArc(THIS->im, cx, cy, w, h, s, e, color);
}

static void gd_resize(INT32 args) {
  int x, y;
  struct object *o;
  gdImagePtr new;

  if( args < 2 )
    error("Too few arguments to gd->resize()\n");

  if( sp[-args].type != T_INT )
    error("Bad argument 1 to gd->resize()\n");
  if( sp[1-args].type != T_INT )
    error("Bad argument 2 to gd->resize()\n");

  x = sp[-args].u.integer;
  y = sp[1-args].u.integer;

  if( THIS->im == NULL )
    error("No Image Open or Created");

  pop_n_elems(args);

/*
  o = clone(file_program, 0)
  (*(gdImagePtr)(o->storage))->im = 

  Work more on this later !
 */
}

/*
 Syntax:
   copy(object source_image, int destX, int destY, int srcX, int srcY,
	int w, int h);
 */
static void gd_copy(INT32 args) {
  int dx, dy, sx, sy, h, w;
  struct object *o;

  if( args < 1 )
    error("Too few arguments to gd->copy()\n");

  if( sp[-args].type != T_OBJECT )
    error("Bad argument 1 to gd->copy()\n");

  o = sp[-args].u.object;

  if( !o->prog || o->prog->inherits[0].prog != gd_program )
    error("Argument 1 to gd->copy() must be a clone of /precompiled/gd\n");

  if( args > 1 ) {
    if( sp[1-args].type != T_INT )
      error("Bad argument 1 to gd->copy()\n");
    dx = sp[1-args].u.integer;
  } else
    dx = 0;

  if( args > 2 ) {
    if( sp[2-args].type != T_INT )
      error("Bad argument 2 to gd->copy()\n");
    dy = sp[2-args].u.integer;
  } else
    dy = 0;

  if( args > 3 ) {
    if( sp[3-args].type != T_INT )
      error("Bad argument 3 to gd->copy()\n");
    sx = sp[3-args].u.integer;
  } else
    sx = 0;

  if( args > 4 ) {
    if( sp[4-args].type != T_INT )
      error("Bad argument 4 to gd->copy()\n");
    sy = sp[4-args].u.integer;
  } else
    sy = 0;

  if( args > 5 ) {
    if( sp[5-args].type != T_INT )
      error("Bad argument 5 to gd->copy()\n");
    w = sp[5-args].u.integer;
  } else
    w = THIS->im->sx;

  if( args > 6 ) {
    if( sp[6-args].type != T_INT )
      error("Bad argument 6 to gd->copy()\n");
    h = sp[6-args].u.integer;
  } else
    h = THIS->im->sy;
 
  gdImageCopy(THIS->im, ((struct gd_image_object *)(o->storage))->im, 
	dx, dy, sx, sy, h, w);
  pop_n_elems(args);

}

static void gd_exact_color(INT32 args) {
  int r, g, b, color;

  if( args < 3 )
    error("Too few arguments to gd->exact_color()\n");

  if( sp[-args].type != T_INT )
    error("Bad argument 1 to gd->exact_color()\n");
  if( sp[1-args].type != T_INT )
    error("Bad argument 2 to gd->exact_color()\n");
  if( sp[3-args].type != T_INT )
    error("Bad argument 3 to gd->exact_color()\n");

  if( THIS->im == NULL )
    error("No Image Open or Created");

  r = sp[-args].u.integer;
  g = sp[1-args].u.integer;
  b = sp[2-args].u.integer;

  pop_n_elems(args);
  color = gdImageColorExact(THIS->im, r, g, b);

  push_int(color);
}

static void gd_closest_color(INT32 args) {
  int r, g, b, color;

  if( args < 3 )
    error("Too few arguments to gd->closest_color()\n");

  if( sp[-args].type != T_INT )
    error("Bad argument 1 to gd->closest_color()\n");
  if( sp[1-args].type != T_INT )
    error("Bad argument 2 to gd->closest_color()\n");
  if( sp[3-args].type != T_INT )
    error("Bad argument 3 to gd->closest_color()\n");

  if( THIS->im == NULL )
    error("No Image Open or Created");

  r = sp[-args].u.integer;
  g = sp[1-args].u.integer;
  b = sp[2-args].u.integer;

  pop_n_elems(args);
  color = gdImageColorClosest(THIS->im, r, g, b);

  push_int(color);
}

  
static void gd_polygon(INT32 args) {
  int color, i, sz;
  gdPointPtr pts;

  if( !args )
    error("Too few arguments to gd->polygon()\n");
  if( sp[-args].type != T_ARRAY )
    error("Bad argument 1 to gd->polygon()\n");

  if( args >= 2 ) {
    if( sp[1-args].type != T_INT )
      error("Bad argument 2 to gd->polygon()\n");
    color = sp[1-args].u.integer;
  } else
    color = THIS->color;

  if( THIS->im == NULL )
    error("No Image Open or Created");

  check_valid_color(color, "polygon");
  THIS->color = color;

  sz = sp[-args].u.array->size;
  if( sz % 2 )
    error("Incorrect number of elems in array passed to gd->polygon()\n");

  pts = (gdPointPtr) malloc((sz / 2 * sizeof(gdPointPtr) + 1));
  
  for( i=0; i<sz/2; i++ ) {
    pts[i].x = (ITEM(sp[-args].u.array)[i*2]).u.integer;
    pts[i].y = (ITEM(sp[-args].u.array)[i*2+1]).u.integer;
  }
  pop_n_elems(args);
  gdImagePolygon(THIS->im, pts, sz/2, color);
}

static void gd_query_transparent(INT32 args) {
  int color;

  if( args )
    error("Too many arguments to gd->query_transparent()\n");

  if( THIS->im == NULL )
    error("No Image Open or Created");

  color = gdImageGetTransparent(THIS->im);
  pop_n_elems(args);
  push_int(color);
}

static void gd_query_interlaced(INT32 args) {
  int color;

  if( args )
    error("Too many arguments to gd->query_interlaced()\n");

  if( THIS->im == NULL )
    error("No Image Open or Created");

  color = gdImageGetInterlaced(THIS->im);
  pop_n_elems(args);
  push_int(color);
}

static void gd_set_interlace(INT32 args) {
  int i;

  if( args < 1 )
    error("Too few arguments to gd->set_interlace()\n");
  
  if( sp[-args].type != T_INT )
    error("Bad argument 1 to gd->set_interlace()\n");

  if( THIS->im == NULL )
    error("No Image Open or Created");

  i = sp[-args].u.integer;
  pop_n_elems(args);
  gdImageInterlace(THIS->im, i);
}

static void gd_tile(INT32 args) {
  struct object *o;
  gdImagePtr tile;

  if( args < 1 )
    error("Too few arguments to gd->tile()\n");
  if( sp[-args].type != T_OBJECT )
    error("Bad argument 1 to gd->tile()\n");

  o = sp[-args].u.object;

  pop_n_elems(args);

  if( THIS->im == NULL )
    error("No Image Open or Created");

  if( !o->prog || o->prog->inherits[0].prog != gd_program )
    error("Argument 1 to gd->tile() must be a clone of /precompiled/gd\n");

  tile = ((struct gd_image_object *)(o->storage))->im;
  if( tile == NULL )
    error("Object passed to gd->tile() does not contain image\n");

  gdImageSetTile(THIS->im, tile);
}

static void gd_set_color(INT32 args) {
  int color;

  if( args < 1 )
    error("Too few arguments to gd->set_color()\n");
  else if( args > 1 )
    error("Too many arguments to gd->set_color()\n");

  if( sp[-args].type != T_INT )
    error("Bad argument 1 to gd->set_color() [expecting int]\n");

  color = sp[-args].u.integer;

  check_valid_color(color, "set_color");
  THIS->color = color;
}

/* Don't know any way other than hard coding this...*/
static void gd_query_fonts(INT32 args) {
  pop_n_elems(args);

  push_string(make_shared_string("tiny"));
  push_string(make_shared_string("small"));
  push_string(make_shared_string("medium"));
  push_string(make_shared_string("large"));
  push_string(make_shared_string("giant"));
  f_aggregate(5);
}

  
gdFontPtr valid_font(char *f) {
  if( !strcasecmp(f, "tiny") )
    return gdFontTiny;
  if( !strcasecmp(f, "small") )
    return gdFontSmall;
  if( !strcasecmp(f, "medium") )
    return gdFontMediumBold;
  if( !strcasecmp(f, "large") )
    return gdFontLarge;
  if( !strcasecmp(f, "giant") )
    return gdFontGiant;
  return (NULL);
}

static void gd_query_font_info(INT32 args) {
  char *c;
  gdFontPtr font;

  if( !args )
    error("Too few arguments to gd->query_font_info()\n");

  if( sp[-args].type != T_STRING )
    error("Bad argument 1 to gd->query_font_info()\n [expecting string]\n");

  c = sp[-args].u.string->str;
  font = valid_font(c);
  pop_n_elems(args);

  if( font == NULL )
    push_int(0);
  else {
    push_int(font->w);
    push_int(font->h);
    f_aggregate(2);
  }    
}
/* Syntax:
void gd->string(string text, int x, int y, string font, int color);
*/
static void gd_string(INT32 args) {
  int color, x, y;
  char *str, *font;
  gdFontPtr fontptr;
 
  if( args < 3 )
    error("Too few arguments to gd->string()\n");

  if( sp[-args].type != T_STRING )
    error("Bad argument 1 to gd->string() [expecting string]\n");
  if( sp[1-args].type != T_INT )
    error("Bad argument 2 to gd->string() [expecting int]\n");
  if( sp[2-args].type != T_INT )
    error("Bad argument 3 to gd->string() [expecting int]\n");

  str = sp[-args].u.string->str;
  x = sp[1-args].u.integer;
  y = sp[2-args].u.integer;

  if( args > 3 ) {
    if( sp[3-args].type != T_STRING )
      error("Bad argument 4 to gd->string() [expecting string]\n");
    font = sp[3-args].u.string->str;
  } else {
    font = (char *) malloc(8);
    strcpy(font, "medium");
  }
  fontptr = valid_font(font);
  if( fontptr == NULL ) {
    char *err;

    err = (char *) malloc(60 + strlen(font));
    sprintf(err, "Unknown font '%s' passed to gd->string()\n", font);
    error(err);
  }

  if( args > 4 ) {
    if( sp[4-args].type != T_INT )
      error("Bad argument 5 to gd->string() [expecting int]\n");
    color = sp[4-args].u.integer;
  } else
    color = THIS->color;

  pop_n_elems(args);

  check_valid_color(color, "string");
  THIS->color = color;

  if( THIS->im == NULL )
    error("No Image Open or Created");

  gdImageString(THIS->im, fontptr, x, y, str, color);
}

/*
**
*/
void init_gd_programs() {
  start_new_program();

  set_init_callback(init_gd_callback);
  set_exit_callback(exit_gd_callback);


  add_function("create", gd_create, "function(int,int:void)", 0);
  add_function("destroy", gd_destroy, "function(:void)", 0);

  add_function("line", gd_line, "function(int,int,int,int,int:void)", 0);
  add_function("point", gd_point, "function(int,int,int:void)", 0);
  add_function("box", gd_box, "function(int,int,int,int,int,int:void)", 0);
  add_function("arc", gd_arc, "function(int,int,int,int,int,int,int:void)", 0);
  add_function("polygon", gd_polygon, "function(array,int:void)", 0);
  add_function("fill", gd_fill, "function(int,int,int:void)", 0);
  add_function("tile", gd_tile, "function(object:void)", 0);

  add_function("alloc_color", gd_alloc_color,
	"function(int,int,int:int)", 0);
  add_function("free_color", gd_free_color, "function(int:void)", 0);
  add_function("set_color", gd_set_color, "function(int:void)", 0);
  add_function("closest_color", gd_closest_color, "function(int,int,int:int)",
	0);
  add_function("exact_color", gd_exact_color, "function(int,int,int:int)", 0);

  add_function("query_color", gd_query_color, "function(int:array)", 0);
  add_function("query_pixel", gd_query_pixel, "function(int,int:int)", 0);

  add_function("print_image", gd_print_image, "function(:void)", 0);
  add_function("create_image", gd_create_image, "function(int,int:int)", 0);
  add_function("load_image", gd_load_image, "function(string,string:int)", 0);

  add_function("query_size", gd_query_size, "function(:array)", 0);
  add_function("resize", gd_resize, "function(int,int:object)", 0);
  add_function("copy", gd_copy, "function(object,int,int,int,int,int,int:void)", 0);

  add_function("set_transparent", gd_set_transparent, "function(int:void)", 0);
  add_function("set_interlace", gd_set_interlace, "function(int:void)", 0);

  add_function("query_transparent", gd_query_transparent, "function(:int)", 0);
  add_function("query_interlaced", gd_query_interlaced, "function(:int)", 0);

  add_function("print_string", gd_string, "function(string,int,int,string,int:void)",
	0);
  add_function("query_fonts", gd_query_fonts, "function(:array)", 0);
  add_function("query_font_info", gd_query_font_info, "function(:array|int)", 0);

  gd_program = end_c_program("/precompiled/gd");
}

void init_gd_efuns() {
}

void exit_gd() {
}

/* To do:
gdBrushed 
gdDashSize 
gdFont - data type
gdFontPtr - data type
gdImage - data type
gdImagePtr -- data type

** gdImageColorAllocate 
** gdImageColorDeallocate 
** gdImageColorTransparent 
** gdImageFilledRectangle 
** gdImageGd 
** gdImageGif 
** gdImageLine 
** gdImageRectangle 
** gdImageBlue 
** gdImageGreen 
** gdImageRed 
** gdImageFill 
** gdImageDestroy 
** gdImageSetPixel 
** gdImageArc 
** gdImageGetPixel 
** gdImageCreate 
** gdImageCreateFromGd 
** gdImageCreateFromGif 
** gdImageCreateFromXbm 
** gdImageColorClosest 
** gdImageColorExact 
** gdImageGetInterlaced 
** gdImageGetTransparent 
** gdImageInterlace 
** gdImageSetTile
** gdImageFilledPolygon 
** gdImagePolygon 

-- gdImageBoundsSafe 
-- gdImageDashedLine 

gdImageCopy 
gdImageCopyResized 
gdImageFillToBorder
gdImageSetBrush 
gdImageSetStyle 
gdMaxColors 

STRING FUNCS:
gdImageChar 
gdImageCharUp 
** gdImageString 
gdImageStringUp 

gdPoint 
gdStyled 
gdStyledBrushed 
gdTiled 
gdTransparent
*/


